MODEL_TYPE_CLF = 'clf'
MODEL_TYPE_REG = 'reg'
MODEL_TYPE_CLU = 'clu'

DATA_TYPE_TABULAR = 'tabular'
DATA_TYPE_TS = 'ts'
DATA_TYPE_IMAGE = 'image'

_GRAPHVIZ_DEFAULT_NODE = {
}

GRAPHVIZ_DEFAULT_NODE_SPLIT = _GRAPHVIZ_DEFAULT_NODE | {
    "shape": "box"
}

GRAPHVIZ_DEFAULT_NODE_LEAF = _GRAPHVIZ_DEFAULT_NODE | {
}